#include "Go.h"

#include <stdio.h>
#include <OS.h>
#include <list>
#include <algorithm>
#include <iostream>

const int diff[4][2] = { {-1,0}, {+1,0}, {0,-1}, {0,+1} };

Go::Go( int width, int height )
:	fWidth(width), fHeight(height),
	fWhiteScore(0), fBlackScore(0),
	fBoard( NULL ),
	fPrintedError(false)
{
	fBoard = new Stone[fWidth*fHeight];
	fTempBoard = new Stone[fWidth*fHeight];
	fTemp2Board = new Stone[fWidth*fHeight];
	
	for ( int y=0; y<fWidth; y++ )
		for ( int x=0; x<fWidth; x++ )
			fBoard[y*fWidth+x] = None;

	fValidPlacements = new bool[fWidth*fHeight];
}

Go::~Go()
{
	delete [] fBoard;
}

bool
Go::ValidPlacement( int x, int y )
{
	return fValidPlacements[y*fWidth+x];
}

bool
Go::CalcValidPlacement( Stone stone, int x, int y )
{
	memcpy( fTemp2Board, fBoard, sizeof(Stone)*GetWidth()*GetHeight() );
	
	bool res = PlaceStone(stone,x,y, false);
	
	memcpy( fBoard, fTemp2Board, sizeof(Stone)*GetWidth()*GetHeight() );
	
	return res;
}

void
Go::UpdateValidPlacements( Stone stone )
{
	for ( int y=0; y<fHeight; y++ )
	{
		for ( int x=0; x<fWidth; x++ )
		{
			fValidPlacements[y*fWidth+x] = CalcValidPlacement(stone,x,y);
		}
	}
}

bool
Go::PlaceStone( Stone stone, int x, int y, bool notify )
{
	if ( StoneAt(x,y) != None )
		return false;
	
	Stone other_stone = None;
	if ( stone == Black )
		other_stone = White;
	if ( stone == White )
		other_stone = Black;
	
	// copy real board to temp so we can mess around with the real one
	memcpy( fTempBoard, fBoard, sizeof(Stone)*GetWidth()*GetHeight() );
	
	// place stone
	fBoard[y*fWidth+x] = stone;
	
	// 
	list<XY> left, right, top, bottom;
	if ( x > 0 && StoneAt(x-1,y) == other_stone )
		GetConnected(x-1,y,left);
	if ( x < GetWidth()-1 && StoneAt(x+1,y) == other_stone )
		GetConnected(x+1,y,right);
	if ( y > 0 && StoneAt(x,y-1) == other_stone )
		GetConnected(x,y-1,top);
	if ( y < GetHeight()-1 && StoneAt(x,y+1) == other_stone )
		GetConnected(x,y+1,bottom);
	
	bool del_left = false, del_right = false, del_top = false, del_bottom = false;
	
	del_left = left.size() > 0 && !ConnectedToFree(left);
	del_right = right.size() > 0 && !ConnectedToFree(right);
	del_top = top.size() > 0 && !ConnectedToFree(top);
	del_bottom = bottom.size() > 0 && !ConnectedToFree(bottom);
	
	if ( del_left )
		SetToNone( left );
	if ( del_right )
		SetToNone( right );
	if ( del_top )
		SetToNone( top );
	if ( del_bottom )
		SetToNone( bottom );
	
	list<XY> conn;
	GetConnected(x,y,conn);
	
	if ( ConnectedToFree(conn) )
	{
		// sanity check - shouldn't really happen...
		if ( fOldStates.size() > 700 )
		{
			if ( !fPrintedError )
			{
				cerr << "Too many states" << endl;
				fPrintedError = true;
			}
			return false;
		}
		
		// now all we have to do is make sure this is not an old state
		bool matching_old = false;
//		for ( list<unsigned char*>::iterator i=fOldStates.begin(); i!=fOldStates.end(); i++ )
		for ( list<Stone*>::iterator i=fOldStates.begin(); i!=fOldStates.end(); i++ )
		{
			//cerr << fOldStates.size() << " old states" << endl;
			
//			if ( memcmp( (*i), fTempBoard, (GetWidth()*GetHeight())/4+1 ) == 0 )
			if ( memcmp( (*i), fBoard, sizeof(Stone)*GetWidth()*GetHeight() ) == 0 )
			{
				matching_old = true;
				break;
			}
		}
		
		if ( !matching_old )
		{ // placement ok.
			if ( notify )
			{
//				unsigned char * state = new unsigned char[(GetWidth()*GetHeight())/4+1];
				Stone * state = new Stone[GetWidth()*GetHeight()];
				
//				memset( state, 0, (GetWidth()*GetHeight())/4+1 );
				
/*				for ( int i=0; i<GetWidth()*GetHeight(); i++ )
				{
					state[i/4] |= (fBoard[i] % 2) << ((i%4)*2);
				}
*/				
				memcpy( state, fBoard, sizeof(Stone)*GetWidth()*GetHeight() );
				
				fOldStates.push_back(state);
				
				NotifyListeners( stone, XY(x,y) );
			}
			
			return true;
		}
	}
	
	// restore old board
	memcpy( fBoard, fTempBoard, sizeof(Stone)*GetWidth()*GetHeight() );
	
	return false;
}

void
Go::GetConnected( int x, int y, list<XY> & connected )
{
	Stone stone = StoneAt(x,y);
	
	// Get list with stones connected to the placed one
	// of the same color, which will be used to check for
	// empty space
	list<XY> visited;
	list<XY> pending;
	
	connected.push_back( XY(x,y) );
	pending.push_back( XY(x,y) );
	
	while ( pending.size() > 0 && connected.size() < (unsigned int)fWidth*fHeight )
	{
		XY xy(pending.front());
		pending.pop_front();
		
		visited.push_back(xy);
		
		for ( int d=0; d<4; d++ )
		{
			XY curr( xy.x+diff[d][0], xy.y+diff[d][1] );
			
			list<XY>::iterator i = find(visited.begin(), visited.end(), curr);
			
			if ( i != visited.end() )
				// already visited
				continue;
			
			i = find(connected.begin(), connected.end(), curr);
			
			if ( i != connected.end() )
				// already connected
				continue;
			
			if ( StoneAt(curr) != stone )
				// wrong stone
				continue;
			
			connected.push_back(curr);
			pending.push_back(curr);
		}
	}
}

bool
Go::ConnectedToFree( list<XY> & connected )
{
	// got list of connected stones. now make sure that at least one
	// of them if connected to a free square
	for ( list<XY>::iterator i = connected.begin(); i != connected.end(); i++ )
	{
		int x = (*i).x, y = (*i).y;
		
		for ( int d=0; d<4; d++ )
		{
			XY curr( x+diff[d][0], y+diff[d][1] );
			if ( StoneAt(curr) == None )
			{
				return true;
			}
		}
	}
	
	return false;
}

bool
Go::ConnectedToFree( int x, int y )
{
	list<XY> connected;
	
	GetConnected( x, y, connected );
	
	return ConnectedToFree( connected );
}

Go::Stone
Go::StoneAt( int x, int y ) const
{
	if ( x < 0 || y < 0 || x >= fWidth || y >= fHeight )
		return OffBoard;
	
	return fBoard[y*fWidth+x];
}

Go::Stone
Go::StoneAt( const XY & xy ) const
{
	if ( xy.x < 0 || xy.y < 0 || xy.x >= fWidth || xy.y >= fHeight )
		return OffBoard;
	
	return fBoard[xy.y*fWidth+xy.x];
}

int
Go::GetWidth() const
{
	return fWidth;
}

int
Go::GetHeight() const
{
	return fHeight;
}

void
Go::UpdateScore()
{
	fWhiteScore = 0;
	fBlackScore = 0;
	
	for ( int y=0; y<GetHeight(); y++ )
	{
		for ( int x=0; x<GetWidth(); x++ )
		{
			switch ( StoneAt(x,y) )
			{
				case White:
					fWhiteScore++;
					break;
					
				case Black:
					fBlackScore++;
					break;
					
				case None:
				{
					list<XY> connected;
					
					GetConnected(x,y,connected);
					
					int white=0, black=0, other=0;
					
					for ( list<XY>::iterator i = connected.begin(); i != connected.end(); i++ )
					{
						int xx = (*i).x, yy = (*i).y;
						
						for ( int d=0; d<4; d++ )
						{
							switch ( StoneAt(xx+diff[d][0], yy+diff[d][1]) )
							{
								case Black:
									black++;
									break;
								case White:
									white++;
									break;
								default:
									other++;
									break;
							}
						}
					}
					
					if ( white == 0 && black > 0 )
						fBlackScore++;
					
					if ( black == 0 && white > 0 )
						fWhiteScore++;
					
					//printf("w: %d, b: %d, o: %d\n", white, black, other);
				}	break;
				
				default:
					break;
			}
		}
	}
}

int
Go::WhiteScore()
{
	if ( fWhiteScore < 0 )
		UpdateScore();
	
	return fWhiteScore;
}

int
Go::BlackScore()
{
	if ( fWhiteScore < 0 )
		UpdateScore();
	
	return fBlackScore;
}

void
Go::SetToNone( list<XY> & l )
{
//	cout << "Clearing " << l.size() << " squares" << endl;
	
	for ( list<XY>::iterator i = l.begin(); i != l.end(); i++ )
	{
		ForceStone( None, *i );
	}
}

void
Go::ForceStone( Stone stone, XY xy )
{
	fBoard[xy.y*fWidth+xy.x] = stone;
}

void
Go::NotifyListeners( Stone color, XY xy )
{
	fWhiteScore = -1;
	
	for ( list< pair<go_callback,void*> >::iterator i=fCallbacks.begin(); i != fCallbacks.end(); i++ )
	{
		(*i).first(this, color, xy, (*i).second);
	}
}

void
Go::AddListener( go_callback cb, void * data )
{
	fCallbacks.push_back( pair<go_callback,void*>(cb,data) );
}
